import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA, KernelPCA
from sklearn.preprocessing import StandardScaler


class PCA_variance:
    def __init__(self, target_variance, max_components=None, **kwargs) -> None:
        self.target_variance = target_variance
        self.max_components = max_components
        self.best = None
        self.pc_decomposition = None
        self.num_components = None
        self.explained_variance = None
        self.kwargs = kwargs

    def run(self, data: pd.DataFrame) -> pd.DataFrame:
        scaler = StandardScaler(with_mean=True, with_std=True)
        x_standarized = pd.DataFrame(scaler.fit_transform(data), columns = data.columns)

        # If explained variance is 0 < expl_var < 1 it computes automatically in function of the explained variance
        self.best = PCA(n_components=self.target_variance if self.max_components is None else self.max_components,
                        random_state=self.kwargs['random_state'])

        values = self.best.fit_transform(x_standarized)

        self.pc_decomposition = pd.DataFrame(self.best.components_,
                                             columns=data.columns,
                                             index = [f'PC-{i + 1}' for i, _ in enumerate(self.best.components_)])
        
        self.num_components = self.best.n_components_
        self.explained_variance = self.best.explained_variance_ratio_
        
        return pd.DataFrame(values, columns=[f'PC-{i}' for i in range(values.shape[1])])

    def transform(self, data):
        return self.best.transform(data)
    
    def inverse_transform(self, data):
        return self.best.inverse_transform(data)
    
    def plot_variance(self):
        sns.lineplot(self.explained_variance, color='#FF6961')
        sns.despine()
        plt.hlines(self.target_variance, 0, len(self.explained_variance), linestyles='dashed', colors='#4D82BC')
        plt.text(len(self.explained_variance) - 40,
                 self.target_variance - (self.target_variance * 0.03),
                 f"Explained variance = {self.target_variance * 100}%",
                 fontfamily='fantasy',
                 color="#4D82BC")
        plt.show()


class PCA_kernel:
    def __init__(self,target_variance, kernel, num_components=None, **kwargs) -> None:
        self.target_variance = target_variance
        self.kernel = kernel
        self.num_components = num_components
        self.eigensol = ['auto','dense','arpack','randomized']
        self.fit_inverse_transform = True
        self.kwargs = kwargs
        
    def run(self, data: pd.DataFrame) -> pd.DataFrame:
        x_norm = pd.DataFrame(StandardScaler().fit_transform(data), columns = data.columns)     
        self.kernel_pca = KernelPCA(n_components = self.num_components, # Num components = None -> Quedarse con todas las que seas > 0
                                    kernel = self.kernel ,
                                    eigen_solver = self.eigensol[0],
                                    fit_inverse_transform = self.fit_inverse_transform)
        
        pca_ft = self.kernel_pca.fit_transform(x_norm)
        self.explained_variance = self.kernel_pca.eigenvalues_/np.sum(self.kernel_pca.eigenvalues_)
        self.vectors = self.kernel_pca.eigenvectors_
        if self.num_components is not None:
            num_comp_exp = next(i for i, s in enumerate([sum(self.explained_variance[:idx+1]) for idx in range(len(self.explained_variance))]) if s > self.target_variance)
        else:
            num_comp_exp = self.num_components
            
        pca_ft_exp = pca_ft[:,:num_comp_exp]
        
        return pd.DataFrame(pca_ft_exp, columns=[f'PC-{i}' for i in range(pca_ft_exp.shape[1])])

    def transform(self, data):
        return self.kernel_pca.transform(data)
    
    def inverse_transform(self, data):
        return self.kernel_pca.inverse_transform(data)
    
    def plot_variance(self):
        sns.lineplot(self.explained_variance, color='#FF6961')
        sns.despine()
        plt.hlines(self.target_variance, 0, len(self.explained_variance), linestyles='dashed', colors='#4D82BC')
        plt.text(len(self.explained_variance) - 40,
                 self.target_variance - (self.target_variance * 0.03),
                 f"Explained variance = {self.target_variance * 100}%",
                 fontfamily='fantasy',
                 color="#4D82BC")
        plt.show()
