import pickle
import numpy as np
import pandas as pd
from typing import Tuple
from sklearn.metrics import mean_absolute_error, mean_absolute_percentage_error


class Ensemble:
    def __init__(self, models) -> None:
        """ Initializes an ensemble of machine learning models.

        Args:
            models (list): A list of machine learning models to be included in the ensemble.
        """
        self.ensemble = models
    
    def predict(self, x) -> np.array:
        """ Generate predictions for the input data using the ensemble.

        Args:
            x (array-like): The input data for making predictions.

        Returns:
            array: The predicted values based on the ensemble's average prediction.

        Notes:
            This method assumes that each model in the ensemble has a `predict` method.
        """
        predictions = []
        for model in self.ensemble:
            predictions.append(model.predict(x))
        
        return predictions
    
    def predict_ensemble(self, x):
        return np.mean(self.predict(x), axis=0)
    
    def score(self, x, y) -> np.array:
        """ Calculate the ensemble's mean score on the given input data and target values.

        Args:
            x (array-like): The input data for scoring.
            y (array-like): The target values for scoring.

        Returns:
            float: The mean score across all models in the ensemble.

        Notes:
            This method assumes that each model in the ensemble has a `score` method.
        """
        scores = []
        for model in self.ensemble:
            scores.append(model.score(x, y))
            
        return np.mean(scores, axis=0)
    
    def compute_mae_mape(self, x, y) -> np.array:
        y_hat = self.predict(x)
        maes = []
        mapes = []
        for pred in y_hat:
            maes.append(mean_absolute_error(y, pred))
            mapes.append(mean_absolute_percentage_error(y, pred))
        
        return maes, mapes
    
    def compute_mae_mape_ensemble(self, x, y):
        maes, mapes = self.compute_mae_mape(x, y)
        return np.mean(maes, axis=0), np.mean(mapes, axis=0)


def load_ensemble(data_name: str, model_data_path: str, seed: int) -> Tuple[pd.DataFrame, Ensemble]: 
    """Return the ensemble of models previously saved

    Args:
        data_name (str): ID of thedata to be loaded
        model_data_path (str): path to models.pickle file

    Returns:
        Ensemble: The set of model on which predictions will be performed
    """
    data_all = pickle.load(open(f'data/all_processed_{seed}.pkl','rb'))
    ensemble = Ensemble(pickle.load(open(model_data_path, 'rb')))
    
    return data_all[data_name], ensemble