import os
from glob import glob
import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import Any
from pickle import dump as pdump

# User imports
from prebio_df import read_imputed
from results_manager import DataProvider, Results
from preprocessing import impute_missing, delete_outliers
from regression import RegressorType, evaluate_regressor, compute_best_params
from dimensionality_reduction import PCA_variance, PCA_kernel
from feature_selection import xgboost_selection, random_forest_selector


class step:
    def __init__(self) -> None:
        self.count = 0
        
    def __call__(self, *args: Any, **kwds: Any) -> int:
        self.count += 1
        return self.count


# ---------------------------------------------
# Main
# ---------------------------------------------
def main(seed=1234):
    ORIGINAL_DF_PATH = './data/polycarbonates_entregable.xlsx'
    VAR_NAMES = ['Mn (kg·mol)-1', 'Mw/Mn', 'conversion']
    EXPLAINED_VARIANCE = 0.9
    
    kwargs = {'random_state': np.random.RandomState(seed=seed), 'n_jobs': -1}
    data_provider = DataProvider()              # Data storage and manager
    results = Results(seed)                     # Result saver and manager
    step_n = step()                             # Counter class
    
    # ---------------------------------------------
    # NaN imputation
    # ---------------------------------------------
    if not os.path.exists('./data/imputed_polycarbonates.xlsx'):
        print(f"\n{'-' * 50}\nStep {step_n()}: NaN Filling\n{'-' * 50}")
        filled_df = impute_missing(ORIGINAL_DF_PATH, **kwargs)  # And adds a column with number of nans per row
        # After saving imputed data remove NaNs column so it doesn't affect in the regression pipeline
        filled_df = filled_df.drop(columns=['NaNs'])
    else:
        filled_df = read_imputed().drop(columns=['NaNs'])
    
    
    # ---------------------------------------------
    # Delete Outliers
    # ---------------------------------------------
    print(f"\n{'-' * 50}\nStep {step_n()}: Outlier Deletion\n{'-' * 50}")
    outlier_indexes = {name.replace('/', ''): [] for name in VAR_NAMES}
    data = {name.replace('/', ''): delete_outliers(filled_df, name, index_holder=outlier_indexes[name.replace('/', '')], **kwargs)
            for name in VAR_NAMES}
    
    # Save outlier indexes!
    with open('./data/outliers.pkl','wb') as f:
        pdump(outlier_indexes, f)
        f.close()
    
    VAR_NAMES[1] = VAR_NAMES[1].replace('/', '')
    
    for name in VAR_NAMES:
        data_provider.set_data(f'RAW_{name}', data[name])
    
    
    # ---------------------------------------------
    # Dimensionality Reduction: PCA / Kernel PCA
    # ---------------------------------------------
    print(f"\n{'-' * 50}\nStep {step_n()}: Dimensionality Reduction\n{'-' * 50}")
    
    pca_var = {name: PCA_variance(EXPLAINED_VARIANCE, **kwargs) for name in VAR_NAMES}
    pca_sample_var = {name: PCA_variance(data[name][0].shape[0], **kwargs) for name in VAR_NAMES}
    
    pca_kernel_poly = {name: PCA_kernel(EXPLAINED_VARIANCE, 'poly',**kwargs) for name in VAR_NAMES}
    pca_kernel_poly_sample = {name: PCA_kernel(data[name][0].shape[0], 'poly',**kwargs) for name in VAR_NAMES}
    
    pca_kernel_rbf = {name: PCA_kernel(EXPLAINED_VARIANCE, 'rbf', **kwargs) for name in VAR_NAMES}
    pca_kernel_rbf_sample = {name: PCA_kernel(data[name][0].shape[0], 'rbf', **kwargs) for name in VAR_NAMES}

    for var_name in tqdm(VAR_NAMES):
        x, y = data_provider.get_data(f'RAW_{var_name}')
        
        new_x = pca_var[var_name].run(x)
        data_provider.set_data(f'RAW_PCA_var_{var_name}', [new_x, y])
        
        new_x = pca_sample_var[var_name].run(x)
        data_provider.set_data(f'RAW_PCA_sam_{var_name}', [new_x, y])
        
        new_x = pca_kernel_poly[var_name].run(x)
        data_provider.set_data(f'RAW_PCA_poly_var_{var_name}', [new_x, y])
        
        new_x = pca_kernel_poly_sample[var_name].run(x)
        data_provider.set_data(f'RAW_PCA_poly_sam_{var_name}', [new_x, y])
        
        new_x = pca_kernel_rbf[var_name].run(x)
        data_provider.set_data(f'RAW_PCA_rbf_var_{var_name}', [new_x, y])
        
        new_x = pca_kernel_rbf_sample[var_name].run(x)
        data_provider.set_data(f'RAW_PCA_rbf_sam_{var_name}', [new_x, y])


    # ---------------------------------------------
    # Feature Selection: XGBoost and Random Forest
    # ---------------------------------------------
    print(f"\n{'-' * 50}\nStep {step_n()}: Feature Selection\n{'-' * 50}")

    for data_name in tqdm(data_provider.ids):
        var_name = data_name.split('_')[-1]
        name = data_name.replace(f'_{var_name}', '')
        
        x, y = data_provider.get_data(data_name)
        y = np.ravel(y)
        
        xgb_vars = xgboost_selection(x, y, tolerance=0.01, **kwargs)
        rf_vars = random_forest_selector(x, y, tolerance=0.01, **kwargs)
        
        data_provider.set_data(f'{name}_XGB_{var_name}', [x[xgb_vars.index], y])
        data_provider.set_data(f'{name}_RF_{var_name}', [x[rf_vars.index], y])
 

    # ---------------------------------------------
    # Regression
    # ---------------------------------------------
    print(f"\n{'-' * 50}\nStep {step_n()}: Regression\n{'-' * 50}")

    if not glob('./configs/*'):
        for name in tqdm(data_provider.ids, desc='Computing best params'):
            d = data_provider.get_data(name)
            compute_best_params(d[0], d[1], name, verbose= 0, **kwargs)

    for name in tqdm(data_provider.ids):
        var_name = name.split('_')[-1]
        row_name = name.replace(f'_{var_name}', '')

        x, y = data_provider.get_data(name)
        
        score_mean, score_std, regressors = evaluate_regressor(RegressorType.LINEAR, x, y, name, **kwargs)
        results.set_value(var_name, row_name, 'LR(mean)', score_mean)
        results.set_value(var_name, row_name, 'LR(std)', score_std)
        results.set_models(f'{name}_linear', regressors)
    
        score_mean, score_std, regressors = evaluate_regressor(RegressorType.SVR, x, y, name, **kwargs)
        results.set_value(var_name, row_name, 'SVR(mean)', score_mean)
        results.set_value(var_name, row_name, 'SVR(std)', score_std)
        results.set_models(f'{name}_svr', regressors)

        score_mean, score_std, regressors = evaluate_regressor(RegressorType.RANDOM_FOREST, x, y, name, **kwargs)
        results.set_value(var_name, row_name, 'RF(mean)', score_mean)
        results.set_value(var_name, row_name, 'RF(std)', score_std)
        results.set_models(f'{name}_forest', regressors)
    
    # ---------------------------------------------
    # SAVE!    
    # ---------------------------------------------
    data_provider.to_pickle(path = f'./data/all_processed_{seed}.pkl')
    results.to_excel()
    results.save_models(path=f'./results/models/{seed}/')


if __name__ == "__main__":
    var_names = ['Mn(kg·mol)-1', 'MwMn', 'conversion']
    seeds = [645, 4578, 72, 2365, 90345, 24, 1859, 1334, 2078, 2446, 7409, 6995, 2041,  449, 9475]
    
    for seed in seeds:
        main(seed=seed)

    for name in var_names:
        dfs = [pd.read_excel(f"./results/regression/{name}_{seed}.xlsx", index_col=0) for seed in seeds]
        mean_df = pd.concat(dfs).groupby(level=0).mean()
        mean_df.to_excel(f"./results/regression/{name.replace(' ','').replace('/','_')}_mean.xlsx")
