import click
import numpy as np
import pandas as pd
from scipy import stats
from sklearn.impute import KNNImputer

from prebio_df import read_prebio
from nan_analysis import missing_values_table, get_nans_per_row

# ---------------------------------------------
# Helper functions
# ---------------------------------------------
def knn_custom_imputation(target: pd.Series, set: pd.DataFrame) -> pd.DataFrame:
    """Impute missing values in the dataset using KNN imputation.
    
    Args:
        target (pd.Series): Target row with missing values to be imputed.
        set (pd.DataFrame): Dataset containing the target row and other rows for imputation.
    
    Returns:
        pd.DataFrame: Dataset with missing values imputed using KNN imputation.
    """
    set.loc[len(set)] = target
    imputer = KNNImputer()
    res = imputer.fit_transform(set)
    
    return pd.DataFrame(res, columns = set.columns)
    
def check_complete_samples(df: pd.DataFrame) -> tuple:
    """Find all samples with 0% of missing values
    
    Args:
        df (pd.DataFrame): Input dataframe
    
    Returns:
        tuple: A tuple containing the indexes of samples with 0% missing values (empty if none),
            an array of NaN counts for each row, and the unique counts of NaNs.
    """
    nans = np.array([r.isnull().sum() for _, r in df.iterrows()])
    indexes = np.where(nans == 0)[0]
    print(f'> The amount of instances with 0% of missing values is {len(indexes)}!')
    unique, counts = np.unique(nans, return_counts=True)
    print(dict(zip(unique, counts)))
    
    return indexes, nans, unique


# ---------------------------------------------
# Main functions
# ---------------------------------------------
def delete_outliers(df: pd.DataFrame, var_name:str, index_holder = [] , dependent_vars = ['Mn (kg·mol)-1', 'Mw/Mn', 'conversion'], **kwargs) -> list:
    """Delete outliers from the dataset using the interquartile range. 

    Args:
        df (pd.DataFrame): _description_
        var_name (str): _description_
        dependent_vars (list, optional): _description_. Defaults to ['Mn (kg·mol)-1', 'Mw/Mn', 'conversion'].

    Returns:
        list: _description_
    """
    print(f"Deleting outliers for {var_name.capitalize()}")
    df_name = df.copy(deep=True)
    index_holder.clear()
    
    # Delete outliers using the interquartile range
    x, y = df_name.drop(dependent_vars, axis=1), pd.DataFrame(df_name[var_name])
    
    # Compute interquantile range and min and max values
    q1 = y.quantile(0.25).iloc[0]
    q3 = y.quantile(0.75).iloc[0]
    iqr = q3 - q1
    
    min_lim = q1 - 1.5 * iqr
    max_lim = q3 + 1.5 * iqr
    
    outliers = np.where((y < min_lim) | (y > max_lim))[0]
    
    # This makes outlier indexes be accesible through index_holder
    for o in outliers:
        index_holder.append(o)
        
    print(f'{len(outliers)} deleted')
    x = x.drop(index=outliers)
    y = np.ravel(y.drop(index=outliers))
    
    return x, y

def impute_missing(ds: str='../data/polycarbonates_entregable.xlsx', df = None, save = True, **kwargs) -> pd.DataFrame:
    """Main function to preprocess the dataset by imputing missing values and storing the result.
    
    Args:
        ds (str): Path to the prebio2 dataset (.xlsx).
    """
    if df is None:
        df = read_prebio(ds)
        df = get_nans_per_row(df)
    
    # Get all the samples with a 0% of missing values
    indexes, nans, unique = check_complete_samples(df)
    full_df = df.iloc[indexes]
    
    # For ease imputing new data we delete categorical columns (None of them contains NaNs)
    full_df = full_df.drop(columns=['Epoxide', 'SMILES', 'Catalyst'])

    # Assert the new dataset is empty
    assert missing_values_table(full_df).empty, "This shouldn't be happening. Your full_df has empty values!"

    # Impute missing data sequentially
    for min_nan in np.sort(unique)[1:]:  # We omit taking samples with 0% of missing values as they will be repeated
        for index in np.where(nans == min_nan)[0]:
            full_df = knn_custom_imputation(df.iloc[index], full_df)

    df.update(full_df)
    missing_values_table(df, title='------ Imputed DataFrame ------')

    if save:
        path = './data/imputed_polycarbonates.xlsx'
        print(f'The new dataset will be stored in {path}')
        df.to_excel(path, index = False)
    
    return df.select_dtypes('number')


@click.command
@click.option('--ds', default='./data/polycarbonates_entregable.xlsx', help='Path to prebio2 dataset (.xlsx)')
def main(ds):
    impute_missing(ds)


if __name__ == "__main__":
    main()
    