import pickle
import tqdm
import numpy as np
import pandas as pd
from enum import Enum
from sklearn.svm import SVR
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.model_selection import LeaveOneOut, cross_validate
from sklearn.metrics import mean_absolute_error, mean_absolute_percentage_error
from load_saved_models import Ensemble


class RegressorType(Enum):
    """An identifier of the regressors we've implemented """
    LINEAR = 1,
    SVR = 2,
    RANDOM_FOREST = 3


# ---------------------------------------------
# Grid Search
# ---------------------------------------------
def scikit_GS(regressor: callable, parameters: dict, X: pd.DataFrame, y: pd.DataFrame, verbose = 0, random_gs = False, **kwargs):
    """ Perform a Grid Search over a set of parameters and return initialized regressor
        Only works with scikit regressors

    Args:
        regressor (callable): scikit regressor callable
        parameters (dict): set of parameters where the grid search is performed
        X (pd.DataFrame): fit x data
        y (pd.DataFrame): fit y data

    Returns:
        The best parameters 
    """
    file_name = kwargs.pop('file_name')
    n_jobs = kwargs.setdefault('n_jobs', -1)
    n_jobs = kwargs.pop('n_jobs')
    
    if random_gs:
        clf = RandomizedSearchCV(
            regressor(**kwargs), parameters,
            n_iter = 100,
            scoring = 'neg_mean_squared_error',
            cv = LeaveOneOut(),
            verbose = verbose,
            n_jobs = n_jobs)
    else: 
        clf = GridSearchCV(regressor(**kwargs), parameters,
                           scoring = 'neg_root_mean_squared_error',
                           cv = LeaveOneOut(),
                           verbose = verbose,
                           n_jobs = n_jobs)
    clf.fit(X, y)
    
    if not random_gs: # Only save if random search is not performed
        print(f'Best parameters for {regressor}: {clf.best_params_}')
        with open(f"./configs/{file_name}.pkl",'wb') as f:
            pickle.dump(clf.best_params_, f)
            f.close()
        
    return clf.best_params_


def find_params(regressor: RegressorType, x, y, verbose = 0, **kwargs):
    """ Trains and performs grid search over a set of regressors (which can be extended in RegressorType) 
    and returns the score.

    Args:
        regressor (RegressorType): _description_
        x (list, np.array): list of train samples
        y (list, np.array): list of train targets
        verbose (int, optional): The level of detail for Grid Search. Defaults to 0.
    """
    random = kwargs.pop('random_state')
    
    if regressor == RegressorType.LINEAR:
        LinearRegression()  # Doesn't have any parameter to tune
    
    if regressor == RegressorType.SVR:
        params = {'kernel': ['poly', 'rbf'],
                  'C': np.arange(0.1, 1.0, 0.3),
                  'epsilon': np.arange(0, 1, 0.1)}

        scikit_GS(SVR, params, x, y, verbose = verbose, **kwargs)
    
    if regressor == RegressorType.RANDOM_FOREST:
        random_params = {'n_estimators': [10, 50, 100, 150, 200],
                         'max_depth': [None] + list(np.random.randint(1, 20, 10)),
                         'min_samples_split': list(np.random.randint(2, 20, 10)),
                         'min_samples_leaf': list(np.random.randint(1, 20, 10)),
                         'max_features': ['sqrt', 'log2', None] + list(np.random.uniform(0.1, 0.9, 10))}
        
        br = scikit_GS(RandomForestRegressor, random_params, x, y, 
                       verbose = verbose, 
                       random_state=random, 
                       random_gs=True,
                       **kwargs)
        
        mss = br['min_samples_split']
        msl = br['min_samples_leaf']
        min_depth = max(2, br['max_depth'] - 2) if br['max_depth'] is not None else 0
        max_depth = max(br['max_depth'] + 3, 5) if br['max_depth'] is not None else 0
        
        new_params = {'n_estimators': [br['n_estimators']],
                      'max_depth': [None] + list(range(min_depth, max_depth)) if br['max_depth'] is not None else [None, 2, 5],
                      'min_samples_split': [max(2, mss - 2), mss, mss + 2],
                      'min_samples_leaf': [max(2, msl - 2), msl, msl + 2],
                      'max_features': [br['max_features']]}
        
        scikit_GS(RandomForestRegressor, new_params, x, y,
                  verbose=verbose,
                  random_state=random,
                  **kwargs)


def compute_best_params(x, y, name: str, verbose = 0, **kwargs):
        
    kwargs['file_name'] = f"SVR_{name}"
    find_params(RegressorType.SVR, x, y, verbose = verbose, **kwargs)

    kwargs['file_name'] = f"RF_{name}"
    kwargs['bootstrap'] = True
    find_params(RegressorType.RANDOM_FOREST, x, y, verbose = verbose, **kwargs)



# ---------------------------------------------
# Regressor Utils
# ---------------------------------------------
def get_regressor(type: RegressorType, name: str, **kwargs) -> callable:
    # Returns the regressor specified with GS loaded parameters
    if type == RegressorType.LINEAR:
        regressor = LinearRegression()
        
    elif type == RegressorType.SVR:
        my_kwargs = pickle.load(open(f"./configs/SVR_{name}.pkl",'rb'))
        regressor = SVR(**my_kwargs)
        
    elif type == RegressorType.RANDOM_FOREST:
        my_kwargs = pickle.load(open(f"./configs/RF_{name}.pkl",'rb'))
        my_kwargs['random_state'] = kwargs['random_state']
        # my_kwargs['n_jobs'] = kwargs['n_jobs']
        regressor = RandomForestRegressor(**my_kwargs)
        
    return regressor


def evaluate_regressor(regressor_type, x, y, name, **kwargs):
    # Load regressor with best params
    regressor = get_regressor(regressor_type, name, **kwargs)

    scores = cross_validate(regressor, x, y,
                            scoring='neg_mean_absolute_error',
                            cv=LeaveOneOut(),
                            n_jobs=kwargs['n_jobs'],
                            return_estimator=True)
    
    ensemble = Ensemble(scores['estimator'])
    score = ensemble.score(x, y)
    
    return np.mean(score), np.std(score), scores['estimator']


def regressor_fit(regressor_type, x, y, name, **kwargs):
    regressor = get_regressor(regressor_type, name, **kwargs)
    loo = LeaveOneOut()
    
    maes = []
    yhat = []
    for train_index, test_index in tqdm.tqdm(loo.split(x)):
        X_train, X_test = x.iloc[train_index], x.iloc[test_index]
        y_train, y_test = y[train_index], y[test_index]

        # Fit the random forest on the training data
        regressor.fit(X_train, y_train)

        # Make predictions on the test data
        y_pred = regressor.predict(X_test)

        # Calculate accuracy and store it
        maes.append(mean_absolute_error(y_test, y_pred))
        yhat.append(float(y_pred))

    mape = mean_absolute_percentage_error(y, yhat)

    return yhat, maes, mape, regressor
